{{old_comment}}
#={{'='*50}}=#
#={{'New Comment'|center(50)}}=#
#={{'Lines starting with "#=" will be ignored!'|center(50)}}=#
#={{'So will empty lines before and after the comment.'|center(50)}}=#
#={{'='*50}}=#
